#include <openvr/openvr_capi.h>
#include <stddef.h>

#define OPENVR_HACK_SPOOF_SENSORS 0
#define OPENVR_HACK_STRICT_POSES 0

extern bool         VR_IsHmdPresent();
extern bool         VR_IsRuntimeInstalled();
extern const char * VR_GetVRInitErrorAsEnglishDescription( EVRInitError error );
extern intptr_t     VR_GetGenericInterface( const char * pchInterfaceVersion, EVRInitError * peError );
extern intptr_t     VR_InitInternal( EVRInitError * peError, EVRApplicationType eType );
extern void         VR_ShutdownInternal();

static EVRInitError                        q2vr_lastError      = EVRInitError_VRInitError_Init_NotInitialized;
static struct VR_IVRSystem_FnTable *       q2vr_fnSystem       = NULL;
static struct VR_IVRCompositor_FnTable *   q2vr_fnCompositor   = NULL;
static struct VR_IVRRenderModels_FnTable * q2vr_fnRenderModels = NULL;
static struct VR_IVROverlay_FnTable *      q2vr_fnOverlay      = NULL;

struct ovrHmdStruct_;
#define HAS_OVR_HMD_STRUCT
#include "header/vr_ovr.h"
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef enum ovrHmdStruct_SessionStatusBits_
{
	SessionStatusBits_IsVisible      = 0x01,
	SessionStatusBits_HmdPresent     = 0x02,
	SessionStatusBits_HmdMounted     = 0x04,
	SessionStatusBits_DisplayLost    = 0x08,
	SessionStatusBits_ShouldQuit     = 0x10,
	SessionStatusBits_ShouldRecenter = 0x20,
	SessionStatusBits_HasInputFocus  = 0x40,
	SessionStatusBits_OverlayPresent = 0x80,
} ovrHmdStruct_SessionStatusBits;

typedef struct ovrHmdStruct_
{
	ovrBool                    Running;
	char                       SessionStatus;
	char                       StringBuffer[32768]; // k_unMaxPropertyStringSize
	ETrackingUniverseOrigin    TrackingOrigin;
	long long                  FrameIndex;
	long long                  StatsIndex;
	Compositor_CumulativeStats BaseStats;

	float    VsyncToPhotons;
	ovrSizei TargetSize;

	ovrEyeRenderDesc RenderDesc[ovrEye_Count];
	ovrHmdDesc       HmdDesc;
	ovrTrackerDesc   TrackerDesc[64]; // k_unMaxTrackedDeviceCount
} ovrHmdStruct;

typedef struct q2vrSessionNode_
{
	struct ovrHmdStruct_      session;
	struct q2vrSessionNode_ * next;
} q2vrSessionNode;
static q2vrSessionNode * q2vr_sessions = NULL;

static ovrMatrix4f ovrMatrix4f_fromHmdMatrix34_t( HmdMatrix34_t M )
{
	ovrMatrix4f result = ovrMatrix4f_Identity();
	int         i, j;
	for( i = 0; i < 3; ++i )
		for( j = 0; j < 4; ++j ) result.M[i][j] = M.m[i][j];
	return result;
}

static ovrResult q2vr_initConvert( EVRInitError error )
{
	switch( error )
	{
		case EVRInitError_VRInitError_None: return ovrSuccess;
		case EVRInitError_VRInitError_Init_InitCanceledByUser:
		case EVRInitError_VRInitError_Init_NoConfigPath:
		case EVRInitError_VRInitError_Init_NoLogPath:
		case EVRInitError_VRInitError_Init_NotAvailableToUtilityApps:
		case EVRInitError_VRInitError_Init_NotInitialized:
		case EVRInitError_VRInitError_Init_NotSupportedWithCompositor:
		case EVRInitError_VRInitError_Init_PathRegistryNotFound:
		case EVRInitError_VRInitError_Init_PathRegistryNotWritable:
		case EVRInitError_VRInitError_Init_SettingsInitFailed:
		case EVRInitError_VRInitError_Init_TooManyObjects:
		case EVRInitError_VRInitError_Init_UserConfigDirectoryInvalid: return ovrError_Initialize;
		case EVRInitError_VRInitError_Init_FileNotFound:
		case EVRInitError_VRInitError_Init_InstallationCorrupt:
		case EVRInitError_VRInitError_Init_InstallationNotFound:
		case EVRInitError_VRInitError_Init_VRClientDLLNotFound: return ovrError_LibLoad;
		case EVRInitError_VRInitError_Init_InvalidInterface: return ovrError_MismatchedAdapters;
		case EVRInitError_VRInitError_Init_HmdNotFound: return ovrError_NoHmd;
		case EVRInitError_VRInitError_Init_AnotherAppLaunching:
		case EVRInitError_VRInitError_Init_AppInfoInitFailed:
		case EVRInitError_VRInitError_Init_NoServerForBackgroundApp:
		case EVRInitError_VRInitError_Init_ShuttingDown: return ovrError_ServerStart;
		case EVRInitError_VRInitError_Init_FactoryNotFound:
		case EVRInitError_VRInitError_Init_InterfaceNotFound: return ovrError_ServiceConnection;
		default: return ovrError_RuntimeException;
	}
}

static void q2vr_UpdateRenderDesc( ovrSession session, ovrEyeType eyeType )
{
	ovrEyeRenderDesc eyeRenderDesc = {};
	ovrFovPort       projection    = {};
	ovrMatrix4f      eyeToHead     = {};

	if( session == NULL ) return;

	q2vr_fnSystem->GetProjectionRaw( eyeType, &projection.LeftTan, &projection.RightTan, &projection.DownTan, &projection.UpTan );
	projection.LeftTan *= -1.0f;
	projection.RightTan *= -1.0f;

	eyeRenderDesc.Eye                         = eyeType;
	eyeRenderDesc.Fov                         = projection;
	eyeRenderDesc.DistortedViewport.Size      = session->TargetSize;
	eyeRenderDesc.DistortedViewport.Pos.x     = eyeType * eyeRenderDesc.DistortedViewport.Size.w;
	eyeRenderDesc.DistortedViewport.Pos.y     = 0;
	eyeRenderDesc.PixelsPerTanAngleAtCenter.x = eyeRenderDesc.DistortedViewport.Size.w * MATH_FLOAT_PIOVER4 / ovrFovPort_GetHorizontalFovRadians( projection );
	eyeRenderDesc.PixelsPerTanAngleAtCenter.y = eyeRenderDesc.DistortedViewport.Size.h * MATH_FLOAT_PIOVER4 / ovrFovPort_GetVerticalFovRadians( projection );

	eyeToHead                              = ovrMatrix4f_fromHmdMatrix34_t( q2vr_fnSystem->GetEyeToHeadTransform( eyeType ) );
	eyeRenderDesc.HmdToEyePose.Orientation = ovrQuatf_fromMatrix4f( eyeToHead );
	eyeRenderDesc.HmdToEyePose.Position    = ovrMatrix4f_GetTranslation( eyeToHead );
	eyeRenderDesc.HmdToEyeOffset           = eyeRenderDesc.HmdToEyePose.Position;
	session->RenderDesc[eyeType]           = eyeRenderDesc;
}

static void q2vr_UpdateHmdDesc( ovrSession session )
{
	ETrackedPropertyError trackedPropertyError = ETrackedPropertyError_TrackedProp_Success;
	ovrHmdDesc            hmdDesc              = {};
	int                   i;

	if( session == NULL ) return;

	hmdDesc.Type = ovrHmd_CV1;
	q2vr_fnSystem->GetStringTrackedDeviceProperty( k_unTrackedDeviceIndex_Hmd, ETrackedDeviceProperty_Prop_ModelNumber_String, hmdDesc.ProductName, 64, &trackedPropertyError );
	q2vr_fnSystem->GetStringTrackedDeviceProperty( k_unTrackedDeviceIndex_Hmd, ETrackedDeviceProperty_Prop_ManufacturerName_String, hmdDesc.Manufacturer, 64, &trackedPropertyError );
	hmdDesc.VendorId  = 0;
	hmdDesc.ProductId = 0;
	q2vr_fnSystem->GetStringTrackedDeviceProperty( k_unTrackedDeviceIndex_Hmd, ETrackedDeviceProperty_Prop_SerialNumber_String, hmdDesc.SerialNumber, 64, &trackedPropertyError );
	hmdDesc.FirmwareMajor         = 0;
	hmdDesc.FirmwareMinor         = 0;
	hmdDesc.AvailableHmdCaps      = 0;
	hmdDesc.DefaultHmdCaps        = 0;
	hmdDesc.AvailableTrackingCaps = ovrTrackingCap_Orientation | ovrTrackingCap_Position;
	hmdDesc.DefaultTrackingCaps   = ovrTrackingCap_Orientation | ovrTrackingCap_MagYawCorrection | ovrTrackingCap_Position;

	if( q2vr_fnSystem->GetBoolTrackedDeviceProperty( k_unTrackedDeviceIndex_Hmd, ETrackedDeviceProperty_Prop_WillDriftInYaw_Bool, &trackedPropertyError ) )
		hmdDesc.AvailableTrackingCaps |= ovrTrackingCap_MagYawCorrection;

	for( i = 0; i < ovrEye_Count; ++i )
	{
		hmdDesc.DefaultEyeFov[i] = session->RenderDesc[i].Fov;
		hmdDesc.MaxEyeFov[i]     = session->RenderDesc[i].Fov;
	}
	hmdDesc.Resolution = session->TargetSize;
	hmdDesc.Resolution.w *= 2;
	hmdDesc.DisplayRefreshRate = q2vr_fnSystem->GetFloatTrackedDeviceProperty( k_unTrackedDeviceIndex_Hmd, ETrackedDeviceProperty_Prop_DisplayFrequency_Float, &trackedPropertyError );
	session->HmdDesc           = hmdDesc;
}

static void q2vr_UpdateTrackerDesc( ovrSession session, unsigned int trackerDescIndex, TrackedDeviceIndex_t index )
{
	ETrackedPropertyError trackedPropertyError = ETrackedPropertyError_TrackedProp_Success;
	ovrTrackerDesc        trackerDesc          = {};

	if( session == NULL ) return;

	if( index == k_unTrackedDeviceIndexInvalid )
	{
		trackerDesc.FrustumHFovInRadians = ovrDegreeToRad( 100.0f );
		trackerDesc.FrustumVFovInRadians = ovrDegreeToRad( 70.0f );
		trackerDesc.FrustumNearZInMeters = 0.4f;
		trackerDesc.FrustumFarZInMeters  = 2.5f;
	}
	else
	{
		trackerDesc.FrustumHFovInRadians = ovrDegreeToRad( q2vr_fnSystem->GetFloatTrackedDeviceProperty( index, ETrackedDeviceProperty_Prop_FieldOfViewLeftDegrees_Float, &trackedPropertyError ) +
		                                                   q2vr_fnSystem->GetFloatTrackedDeviceProperty( index, ETrackedDeviceProperty_Prop_FieldOfViewRightDegrees_Float, &trackedPropertyError ) );
		trackerDesc.FrustumVFovInRadians = ovrDegreeToRad( q2vr_fnSystem->GetFloatTrackedDeviceProperty( index, ETrackedDeviceProperty_Prop_FieldOfViewTopDegrees_Float, &trackedPropertyError ) +
		                                                   q2vr_fnSystem->GetFloatTrackedDeviceProperty( index, ETrackedDeviceProperty_Prop_FieldOfViewBottomDegrees_Float, &trackedPropertyError ) );
		trackerDesc.FrustumNearZInMeters = q2vr_fnSystem->GetFloatTrackedDeviceProperty( index, ETrackedDeviceProperty_Prop_TrackingRangeMinimumMeters_Float, &trackedPropertyError );
		trackerDesc.FrustumFarZInMeters  = q2vr_fnSystem->GetFloatTrackedDeviceProperty( index, ETrackedDeviceProperty_Prop_TrackingRangeMaximumMeters_Float, &trackedPropertyError );
	}
	session->TrackerDesc[trackerDescIndex] = trackerDesc;
}

ovrDetectResult ovr_Detect( int timeout )
{
	ovrDetectResult result        = {};
	result.IsOculusServiceRunning = VR_IsRuntimeInstalled() ? ovrTrue : ovrFalse;
	result.IsOculusHMDConnected   = VR_IsHmdPresent() ? ovrTrue : ovrFalse;
	return result;
}

ovrEyeRenderDesc ovr_GetRenderDesc( ovrSession session, ovrEyeType eyeType, ovrFovPort fov )
{
	ovrEyeRenderDesc eyeRenderDesc = {};
	if( session != NULL && eyeType < ovrEye_Count )
	{
		eyeRenderDesc                             = session->RenderDesc[eyeType];
		eyeRenderDesc.Fov                         = fov;
		eyeRenderDesc.PixelsPerTanAngleAtCenter.x = eyeRenderDesc.DistortedViewport.Size.w / ( fov.LeftTan + fov.RightTan );
		eyeRenderDesc.PixelsPerTanAngleAtCenter.y = eyeRenderDesc.DistortedViewport.Size.h / ( fov.UpTan + fov.DownTan );
	}
	return eyeRenderDesc;
}

ovrHmdDesc ovr_GetHmdDesc( ovrSession session )
{
	ovrHmdDesc hmdDesc = {};
	if( session != NULL )
		hmdDesc = session->HmdDesc;
	else
		hmdDesc.Type = VR_IsHmdPresent() ? ovrHmd_CV1 : ovrHmd_None;
	return hmdDesc;
}

ovrResult ovr_Create( ovrSession * pSession, ovrGraphicsLuid * pLuid )
{
	ETrackedPropertyError trackedPropertyError = ETrackedPropertyError_TrackedProp_Success;
	TrackedDeviceIndex_t  trackerIndices[64]   = {k_unTrackedDeviceIndexInvalid}; // k_unMaxTrackedDeviceCount
	q2vrSessionNode *     node                 = NULL;
	uint32_t              trackerCount         = 0;
	uint64_t              luid                 = 0;
	int                   i;

	if( pSession == NULL ) return ovrError_InvalidParameter;
	*pSession = NULL;

	node = (q2vrSessionNode *)malloc( sizeof( q2vrSessionNode ) );
	if( node == NULL ) return ovrError_MemoryAllocationFailure;

	node->next            = q2vr_sessions;
	node->session.Running = ovrTrue;

	node->session.SessionStatus = SessionStatusBits_HmdMounted;
	if( VR_IsHmdPresent() ) node->session.SessionStatus |= SessionStatusBits_HmdPresent;
	if( q2vr_fnSystem->IsInputAvailable() ) node->session.SessionStatus |= SessionStatusBits_HasInputFocus;
	if( q2vr_fnOverlay->IsDashboardVisible() ) node->session.SessionStatus |= SessionStatusBits_OverlayPresent;

	node->session.TrackingOrigin = ETrackingUniverseOrigin_TrackingUniverseSeated;
	node->session.FrameIndex     = 0;
	node->session.StatsIndex     = 0;

	q2vr_fnSystem->GetRecommendedRenderTargetSize( (uint32_t *)&node->session.TargetSize.w, (uint32_t *)&node->session.TargetSize.h );
	node->session.VsyncToPhotons = q2vr_fnSystem->GetFloatTrackedDeviceProperty( k_unTrackedDeviceIndex_Hmd, ETrackedDeviceProperty_Prop_SecondsFromVsyncToPhotons_Float, &trackedPropertyError );

	for( i = 0; i < ovrEye_Count; ++i ) q2vr_UpdateRenderDesc( &node->session, i );
	q2vr_UpdateHmdDesc( &node->session );

	trackerCount = q2vr_fnSystem->GetSortedTrackedDeviceIndicesOfClass( ETrackedDeviceClass_TrackedDeviceClass_TrackingReference, trackerIndices, k_unMaxTrackedDeviceCount, -1 );
#if OPENVR_HACK_SPOOF_SENSORS
	if( trackerCount < 3 ) trackerCount = 3;
#endif
	for( i = 0; i < trackerCount; ++i ) q2vr_UpdateTrackerDesc( &node->session, i, trackerIndices[i] );

	q2vr_fnSystem->GetOutputDevice( &luid, ETextureType_TextureType_OpenGL, NULL );
	if( pLuid != NULL && luid != 0 ) memcpy( pLuid, &luid, sizeof( ovrGraphicsLuid ) );

#if OPENVR_HACK_STRICT_POSES
	q2vr_fnCompositor->SetTrackingSpace( node->session.TrackingOrigin );
#endif
	if( ovrMatrix4f_Distance( ovrMatrix4f_fromHmdMatrix34_t( q2vr_fnSystem->GetSeatedZeroPoseToStandingAbsoluteTrackingPose() ), ovrMatrix4f_Identity() ) < FLT_EPSILON )
		q2vr_fnSystem->ResetSeatedZeroPose();

	*pSession     = &node->session;
	q2vr_sessions = node;
	return ovrSuccess;
}

ovrResult ovr_GetInputState( ovrSession session, ovrControllerType controllerType, ovrInputState * inputState )
{
	ovrInputState state = {};

	if( inputState == NULL ) return ovrError_InvalidParameter;
	if( session == NULL ) return ovrError_InvalidSession;

	// TODO InputManager

	*inputState = state;
	return ovrSuccess;
}

ovrResult ovr_GetSessionStatus( ovrSession session, ovrSessionStatus * sessionStatus )
{
	static ovrBool first_call = ovrTrue;
	char           status     = 0;

	if( sessionStatus == NULL ) return ovrError_InvalidParameter;
	if( session == NULL ) return ovrError_InvalidSession;

	// TODO: This is gross. Looks like a hack. Remove?
	sessionStatus->IsVisible = q2vr_fnCompositor->CanRenderScene() && !first_call;
	first_call               = ovrFalse;

	status                        = session->SessionStatus;
	sessionStatus->HmdPresent     = status & SessionStatusBits_HmdPresent;
	sessionStatus->HmdMounted     = status & SessionStatusBits_HmdMounted;
	sessionStatus->DisplayLost    = status & SessionStatusBits_DisplayLost;
	sessionStatus->ShouldQuit     = status & SessionStatusBits_ShouldQuit;
	sessionStatus->ShouldRecenter = status & SessionStatusBits_ShouldRecenter;
	//	sessionStatus->HasInputFocus  = status & SessionStatusBits_HasInputFocus;
	//	sessionStatus->OverlayPresent = status & SessionStatusBits_OverlayPresent;
	//	sessionStatus->DepthRequested = ovrFalse;
	return ovrSuccess;
}

ovrResult ovr_Initialize( const ovrInitParams * params )
{
	char tableName[128] = {};

	if( q2vr_lastError == EVRInitError_VRInitError_None ) return ovrSuccess;
	VR_InitInternal( &q2vr_lastError, EVRApplicationType_VRApplication_Scene );

	if( q2vr_lastError == EVRInitError_VRInitError_None )
	{
		sprintf( tableName, "FnTable:%s", IVRSystem_Version );
		q2vr_fnSystem = (struct VR_IVRSystem_FnTable *)VR_GetGenericInterface( tableName, &q2vr_lastError );
	}
	if( q2vr_lastError == EVRInitError_VRInitError_None )
	{
		sprintf( tableName, "FnTable:%s", IVRCompositor_Version );
		q2vr_fnCompositor = (struct VR_IVRCompositor_FnTable *)VR_GetGenericInterface( tableName, &q2vr_lastError );
	}
	if( q2vr_lastError == EVRInitError_VRInitError_None )
	{
		sprintf( tableName, "FnTable:%s", IVRRenderModels_Version );
		q2vr_fnRenderModels = (struct VR_IVRRenderModels_FnTable *)VR_GetGenericInterface( tableName, &q2vr_lastError );
	}
	if( q2vr_lastError == EVRInitError_VRInitError_None )
	{
		sprintf( tableName, "FnTable:%s", IVROverlay_Version );
		q2vr_fnOverlay = (struct VR_IVROverlay_FnTable *)VR_GetGenericInterface( tableName, &q2vr_lastError );
	}
	return q2vr_initConvert( q2vr_lastError );
}

ovrResult ovr_RecenterTrackingOrigin( ovrSession session )
{
	if( session == NULL ) return ovrError_InvalidSession;
	q2vr_fnSystem->ResetSeatedZeroPose();
	return ovrSuccess;
}

ovrResult ovr_SetTrackingOriginType( ovrSession session, ovrTrackingOrigin origin )
{
	if( session == NULL ) return ovrError_InvalidSession;
	session->TrackingOrigin = origin;
#if OPENVR_HACK_STRICT_POSES
	q2vr_fnCompositor->SetTrackingSpace( session->TrackingOrigin );
#endif
	return ovrSuccess;
}

ovrTrackerDesc ovr_GetTrackerDesc( ovrSession session, unsigned int trackerDescIndex )
{
	ovrTrackerDesc trackerDesc = {};
	if( session != NULL && trackerDescIndex < k_unMaxTrackedDeviceCount ) trackerDesc = session->TrackerDesc[trackerDescIndex];
	return trackerDesc;
}

ovrTrackingState ovr_GetTrackingState( ovrSession session, double absTime, ovrBool latencyMarker )
{
	ovrTrackingState trackingState = {};
	if( session == NULL ) return trackingState;
	// TODO InputManager
	return trackingState;
}

double ovr_GetPredictedDisplayTime( ovrSession session, long long frameIndex )
{
	double prediction = ovr_GetTimeInSeconds();
	if( session == NULL || session->FrameIndex == 0 ) return prediction;
	prediction += q2vr_fnCompositor->GetFrameTimeRemaining();
	prediction += session->VsyncToPhotons;

	if( frameIndex > 0 )
		prediction += (double)( frameIndex - session->FrameIndex ) / session->HmdDesc.DisplayRefreshRate;
	else
		prediction += 1.0 / session->HmdDesc.DisplayRefreshRate;
	return prediction;
}

unsigned int ovr_GetConnectedControllerTypes( ovrSession session )
{
	// TODO InputManager
	return ovrControllerType_None;
}

void ovr_Destroy( ovrSession session )
{
	q2vrSessionNode * current = q2vr_sessions;
	q2vrSessionNode * last    = NULL;

	while( current != NULL )
	{
		if( &current->session == session )
		{
			if( last != NULL ) last->next = current->next;
			if( q2vr_sessions == current ) q2vr_sessions = current->next;
			free( current );
		}
		last    = current;
		current = current->next;
	}
}

void ovr_GetLastErrorInfo( ovrErrorInfo * errorInfo )
{
	if( errorInfo == NULL ) return;
	errorInfo->ErrorString[511] = 0;
	strncpy( errorInfo->ErrorString, VR_GetVRInitErrorAsEnglishDescription( q2vr_lastError ), 511 );
	errorInfo->Result = q2vr_initConvert( q2vr_lastError );
}

void ovr_Shutdown()
{
	while( q2vr_sessions != NULL ) ovr_Destroy( &q2vr_sessions->session );
	VR_ShutdownInternal();
	q2vr_lastError      = EVRInitError_VRInitError_Init_NotInitialized;
	q2vr_fnSystem       = NULL;
	q2vr_fnCompositor   = NULL;
	q2vr_fnRenderModels = NULL;
}
